@@ -13,8 +13,8 @@ def _split_what(what):
     Returns a tuple of `frozenset`s of classes and attributes.
     """
     return (
-        frozenset(cl for cl in what if isclass(cl)),
-        frozenset(cl for cl in what if isinstance(cl, Attribute)),
+        frozenset(cls for cls in what if isclass(cls)),
+        frozenset(cls for cls in what if isinstance(cls, Attribute)),
     )
 
 
